def build_pkt_plot_dict(t, w, y0, y1, c, ids):
    import numpy as np
    d = dict()

    if(type(t) is not list and type(t) is not tuple):
        raise Exception("ERROR: input t must be a list/tuple of np arrays")
    
    if(len(t) != len(w) or len(t) != len(y0) or len(t) != len(y1) or len(t) != len(c)):
        raise Exception("ERROR: inputs must be iterables of the same length")

    for idx,(ti, wi, y0i, y1i, ci, idsi) in enumerate(zip(t,w,y0,y1,c,ids)):
        if(np.isscalar(wi)):
            arr_w = wi*np.ones_like(ti)
        else:
            arr_w = wi
        
        if(np.isscalar(y0i)):
            arr_y0 = y0i*np.ones_like(ti)
        else:
            arr_y0 = y0i
    
        if(np.isscalar(y1i)):
            arr_y1 = y1i*np.ones_like(ti)
        else:
            arr_y1 = y1i
    
        if(np.isscalar(ci)):
            arr_c = ci*np.ones_like(ti)
        else:
            arr_c = ci

        if(np.isscalar(idsi)):
            arr_ids = idsi*np.ones_like(ti)
        else:
            arr_ids = idsi

        if(idx == 0):
            all_t   = np.array(ti)
            all_w   = np.array(arr_w)
            all_y0  = np.array(arr_y0)
            all_y1  = np.array(arr_y1)
            all_c   = np.array(arr_c)
            all_ids = np.array(arr_ids)
        else:
            all_t   = np.concatenate( (all_t,  np.array(ti)) )
            all_w   = np.concatenate( (all_w,  np.array(arr_w)) )
            all_y0  = np.concatenate( (all_y0, np.array(arr_y0)) )
            all_y1  = np.concatenate( (all_y1, np.array(arr_y1)) )
            all_c   = np.concatenate( (all_c,  np.array(arr_c)) )
            all_ids = np.concatenate( (all_ids,  np.array(arr_ids)) )

    #Sort the output arrays 
    idx_sort = np.argsort(all_t)
    all_t   = all_t[idx_sort]
    all_w   = all_w[idx_sort]
    all_y0  = all_y0[idx_sort]
    all_y1  = all_y1[idx_sort]
    all_c   = all_c[idx_sort]
    all_ids = all_ids[idx_sort]
    
    d['p_t']  = all_t.astype('uint64')
    d['p_w']  = all_w.astype('uint16')
    d['p_y0'] = all_y0.astype('uint16')
    d['p_y1'] = all_y1.astype('uint16')
    d['p_c']  = all_c.astype('uint32')  
    d['p_id'] = all_ids.astype('uint64')
    
    return d

def get_t_offset_vec(tx_low_ap, rx_sta):
    import numpy as np
    import pandas as pd
    
    # Line up Tx Low and Rx timestamps
    #beacon_tx = tx_low_ap[(tx_low_ap['pkt_type'] == 11) & (tx_low_ap['addr2'] == addr_ap)]
    beacon_tx = tx_low_ap[(tx_low_ap['pkt_type'] == 11)]
    beacon_rx_sta = rx_sta[(rx_sta['pkt_type'] == 11) & (rx_sta['addr2'] == tx_low_ap['addr2'][0]) & (rx_sta['fcs_result'] == 0)]
    
    t_diff_vec = []
    t_diff_vec_ts = []
    for idx in range(len(beacon_rx_sta)):
        temp_beacon_rx = beacon_rx_sta[idx]    
        seq_num = temp_beacon_rx['mac_seq']
        temp_beacon_tx = beacon_tx[beacon_tx['mac_seq'] == seq_num]    
        
        if len(temp_beacon_tx) > 0:
            t_diff_temp = []
            t_diff_temp = np.insert(t_diff_temp, 0, np.int64(temp_beacon_rx['timestamp'])-np.int64(temp_beacon_tx['timestamp']))            
            
            t_diff = t_diff_temp[np.abs(t_diff_temp) == np.min(np.abs(t_diff_temp))] - 24
            t_diff_ts = temp_beacon_rx['timestamp']
                
            if np.abs(t_diff) < 100000:
                t_diff_vec = np.insert(t_diff_vec, 0, t_diff)
                t_diff_vec_ts = np.insert(t_diff_vec_ts, 0, t_diff_ts)
    t_diff_vec = t_diff_vec[::-1]
    t_diff_vec_ts = t_diff_vec_ts[::-1]
    
    t_diff_vec_ts_pd = pd.to_datetime(t_diff_vec_ts, unit='us')
    t_diff_vec_pd = pd.Series(t_diff_vec, index=t_diff_vec_ts_pd)
    
    return t_diff_vec_pd


def proc_txrx_log_to_np(log_filename):
    import wlan_exp.log.util as log_util
    import wlan_exp.log.util_hdf as hdf_util

    entries_filt = ['NODE_INFO', 'RX_OFDM', 'TX', 'TX_LOW', 'TIME_INFO']

    log_data  = hdf_util.hdf5_to_log_data(filename=log_filename)
    log_index = log_util.filter_log_index(hdf_util.hdf5_to_log_index(filename=log_filename), 
                                            include_only=entries_filt,
                                             merge={'RX_OFDM': ['RX_OFDM', 'RX_OFDM_LTG'], 
                                             'TX'     : ['TX', 'TX_LTG'],
                                             'TX_LOW' : ['TX_LOW', 'TX_LOW_LTG']})
    log_np = log_util.log_data_to_np_arrays(log_data,  log_index)

    return log_np


def proc_txrx_logs_to_flows(nodes, flows, flow_conds_common=[]):
    """Processes raw log files from multiple nodes to extract Tx and Rx
    events for an arbitrary number of flows. The arguments must be:
    
    nodes: list of node descriptions. Each entry in the nodes list must be a
        2-tuple (node_name, HDF5_raw_log_file_name)

    flows: list of flow descriptions. Each entry in the flows list must be a
        4-tuple (src, dest, conds_tx, conds_rx), where:
          src/dest: either string name matching name in nodes list or MAC address as uint64
          conds: list of 3-tuples (field_name, operator, value); use [] for no conditions

    flow_conds_common: list of flow matching conditions, applied to Tx/Rx of every flow
      Pass [] (empty list) for no conditions
      Each entry in the flow_conds_common list must be a 3-tuple:
        (field_name, operator, value)

    """

    import numpy as np
    import pandas as pd
        
    #Extract the numpy arrays for each log entry type
    all_entries_np = [proc_txrx_log_to_np(f) for (n,f) in nodes]
    all_tx_l = [log['TX_LOW'] for log in all_entries_np]
    all_rx   = [log['RX_OFDM'] for log in all_entries_np]

    for ref_idx,node_tx in enumerate(all_tx_l):
        if(sum(node_tx['pkt_type'] == 11)):
            #This node had management Tx - use it as the time reference
            tx_low_time_ref = node_tx
            break
    
    for idx,(node_tx, node_rx) in enumerate(zip(all_tx_l,all_rx)):
        if(idx == ref_idx):
            #Don't correct times of reference node
            continue

        t_diff_pd = get_t_offset_vec(tx_low_time_ref, node_rx)

        t_pd = pd.to_datetime(node_rx['timestamp'], unit='us')
        t_pd = pd.Series(t_pd, index=t_pd) 
        t_diff_pd_rs = t_diff_pd.reindex(t_pd.index.union(t_diff_pd.index))
        t_diff_pd_rs = t_diff_pd_rs.interpolate(method='time')
        t_diff_pd_rs = t_diff_pd_rs.reindex(t_pd.index)
        t_corr = np.round(t_diff_pd_rs.fillna(value=0).values)
        
        node_rx['timestamp'] = node_rx['timestamp'] - t_corr

        t_pd = pd.to_datetime(node_tx['timestamp'], unit='us')
        t_pd = pd.Series(t_pd, index=t_pd) 
        t_diff_pd_rs = t_diff_pd.reindex(t_pd.index.union(t_diff_pd.index))
        t_diff_pd_rs = t_diff_pd_rs.interpolate(method='time')
        t_diff_pd_rs = t_diff_pd_rs.reindex(t_pd.index)
        t_corr = np.round(t_diff_pd_rs.fillna(value=0).values)
        
        node_tx['timestamp'] = node_tx['timestamp'] - t_corr

    #Extract each node's MAC address from the NODE_INFO entry in each log
    node_addrs = [log['NODE_INFO']['wlan_mac_addr'][0] for log in all_entries_np]
    
    #%%
    ## Flow Extraction
    flows_tx = list()
    flows_rx = list()
    
    #Iterate over flows
    for f_idx,flow in enumerate(flows):
    
        f_src = flow[0]
        f_dst = flow[1]
        f_conds_tx = flow[2]
        f_conds_rx = flow[3]
        
        #Find the MAC address for the src/dest nodes
        if(type(f_src) is str):
            addr_src = node_addrs[ [n[0] for n in nodes].index(f_src) ]
        else:
            addr_src = f_src
    
        if(type(f_dst) is str):
            addr_dst = node_addrs[ [n[0] for n in nodes].index(f_dst) ]
        else:
            addr_dst = f_dst
            
        ## FLOW TX
        #Extract TX_LOW entries for the source node
        flow_src_tx_l = all_tx_l[ [n[0] for n in nodes].index(f_src) ]
    
        #Filter TX_LOW entries to isolate packets for this flow
        flow_tx_idx =  ((flow_src_tx_l['addr1'] == addr_dst) &
                        (flow_src_tx_l['addr2'] == addr_src))

        for (field,op,val) in f_conds_tx:
            flow_tx_idx = flow_tx_idx & (eval("flow_src_tx_l['{0}'] {1} {2}".format(field,op,val)))

        for (field,op,val) in flow_conds_common:
            flow_tx_idx = flow_tx_idx & (eval("flow_src_tx_l['{0}'] {1} {2}".format(field,op,val)))
    
        #Index of flows_tx is index of flow in flows list above
        flows_tx.insert(f_idx, flow_src_tx_l[flow_tx_idx])
    
        ## FLOW RX
        #Iterate over potential receivers for this flow
        flows_rx.insert(f_idx, list())
        for n_idx,node in enumerate(nodes):
    
            n_name = node[0]
            if(n_name == f_src):
                #Skip this flow's source node
                flows_rx[f_idx].insert(n_idx, [])
                continue
            
            #This node's Rx entries
            node_rx = all_rx[n_idx]
            
            #Filter RX_OFDM entries to isolate packets for this flow
            flow_rx_idx = ((node_rx['addr1'] == addr_dst) & 
                           (node_rx['addr2'] == addr_src))
    
            for (field,op,val) in f_conds_rx:
                flow_rx_idx = flow_rx_idx & (eval("node_rx['{0}'] {1} {2}".format(field,op,val)))

            for (field,op,val) in flow_conds_common:
                flow_rx_idx = flow_rx_idx & (eval("node_rx['{0}'] {1} {2}".format(field,op,val)))

           # print(f_idx, n_idx, hex(addr_dst), hex(addr_src), sum(flow_rx_idx))
            
            #flows_rx is 2D list: [flow_index][rx_node_index]
            flows_rx[f_idx].insert(n_idx, node_rx[flow_rx_idx])

    return (flows_tx, flows_rx, all_entries_np)
    
